function PvEFrames_DefaultSettings()
  PvEFrames["width"]=100
  PvEFrames["height"]=24
  PvEFrames["columns"]=2
  PvEFrames["rows"]=4
  PvEFrames["castBar"]=1
  PvEFrames["locked"]=1
  PvEFrames["show"]=1
  PvEFrames["shortenName"]=1
  PvEFrames["settingsVersion"]=1
end

if PvEFrames==nil then
  PvEFrames={}
  PvEFrames_DefaultSettings()
end






--These are functions for the options frame

function WidthChanged(self, arg1)
  PvEFrames["width"]=arg1
  PvEFrames_AdjustFrames()
end

function HeightChanged(self, arg1)
  PvEFrames["height"]=arg1
  PvEFrames_AdjustFrames()
end

function RowsChanged(self, arg1)
  PvEFrames["rows"] = arg1
  PvEFrames_AdjustFrames()
end

function ColumnsChanged(self, arg1)
  PvEFrames["columns"] = arg1
  PvEFrames_AdjustFrames()
end

function PvEFramesToggleLock(value)
  if value==1 then
    PvEFrameMoveButton:Hide()
  else
    PvEFrameMoveButton:Show()
  end
  if PvEFrames["locked"]==1 then PvEFramesLockButton:SetNormalTexture("Interface\\AddOns\\PvEFrames\\textures\\lock"); else PvEFramesLockButton:SetNormalTexture("Interface\\AddOns\\PvEFrames\\textures\\unlock"); end
end





--The actual load function

function PvEFrames_OnLoad(self)
  --Previously this function created 8 frames and placed them where they needed to go, but with the new 1-64 frame option
  --I will be creating all 64 and then arranging and showing frames as needed.
  for i=1, 64 do
    local f=CreateFrame("Frame","PvEFrame"..i,PvEFrameBackground,"PvEFrameTemplate",i)
    local g=CreateFrame("Button","PvEFrameButton"..i,PvEFrameBackground,"PvETargetButtonTemplate",i)
  end
end

function PvEFrames_OnEnterWorld(self)
  if PvEFrames["settingsVersion"]~=1 then
    PvEFrames=nil
    PvEFrames={}
    PvEFrames_DefaultSettings()
  end
  PvEFrames_AdjustFrames()
  PvEFrames_EnterWorld(self)
end

function PvEFrames_AdjustFrames()
  local width, height, rows, columns, castBar = PvEFrames["width"], PvEFrames["height"], PvEFrames["rows"], PvEFrames["columns"], PvEFrames["castBar"]
  PvEFrameBackground:SetSize(width*columns+4,height*rows+4)
  for i=1, 64 do
    _G["PvEFrame"..i]:Hide()
    _G["PvEFrameButton"..i]:Hide()
  end
  for i=1, (rows*columns) do
    local r = mod(i,rows)
    if r==0 then
      r = rows
    end
    local c = floor((i-1)/rows)+1
    local f, g = _G["PvEFrame"..i], _G["PvEFrameButton"..i]
    f:SetSize(width,height)
    g:SetSize(width,height)
    if castBar==1 then
      f.health:SetHeight(2/3*height+1)
      f.cast:SetHeight(height/3+1)
      f.cast:Show()
    else
      f.health:SetHeight(height)
      f.cast:Hide()
    end
    f:SetPoint("TOPLEFT",PvEFrameBackground,"TOPLEFT",(c-1)*width+2,-(r-1)*height-2)
    g:SetPoint("TOPLEFT",PvEFrameBackground,"TOPLEFT",(c-1)*width+2,-(r-1)*height-2)
    f:Show()
    g:Show()
  end
end

function PvEFrames_EnterWorld(self)
  local instance = IsInInstance()
  if instance==1 then
    self:Show()
  elseif PvEFrames["show"]==1 then
    self:Show()
  else
    self:Hide()
  end
end

function PvEFrames_Visibility(value)
  local instance = IsInInstance()
  if value==1 then
    PvEFrameBackground:Show()
  elseif value==nil and instance==1 then
    PvEFrameBackground:Show()
  else
    PvEFrameBackground:Hide()
  end
end














--These functions are for changing the actual appearance of the bars based on the given data.

function SetName(frame, name)
  if PvEFrames["shortenName"]==1 then
    name = ShortenName(name)
  end
  frame.health.bar.name:SetText(name)
  frame.health.bar.name:Show()
  frame.health.bar.icon:Hide()
end

function SetBorderColor(frame,color)
  frame.health:SetBackdropBorderColor(color,color,color)
  frame.cast:SetBackdropBorderColor(color,color,color)
end

function SetHealthBarValues(frame, health, max)
  frame.health.bar:SetMinMaxValues(0,max)
  frame.health.bar:SetValue(health)
end

function SetHealthBarColor(frame, unit)
  --local _, _, _, rawthreatpct = UnitDetailedThreatSituation("player", unit)
  local status = UnitThreatSituation("player", unit)
  if status then
    local a, b, c = HEALTH_BAR_COLORS[status][1], HEALTH_BAR_COLORS[status][2], HEALTH_BAR_COLORS[status][3]
    frame.health.bar:SetStatusBarColor(a,b,c)
  end
end


HEALTH_BAR_COLORS = {
	[0] = {.69,.69,.69},
	[1] = {1,1,.47},
	[2] = {1,.6,0},
	[3] = {1,0,0}
}

function SetRaidIcon(frame, icon)
  local a = RAID_ICON_COORDS[icon][1]
  local b = RAID_ICON_COORDS[icon][2]
  local c = RAID_ICON_COORDS[icon][3]
  local d = RAID_ICON_COORDS[icon][4]
  frame.health.bar.icon:SetTexCoord(a,b,c,d)
  frame.health.bar.icon:Show()
  frame.health.bar.name:Hide()
end

RAID_ICON_COORDS = {
	[1] = {0,.25,0,.25},
	[2] = {.25,.5,0,.25},
	[3] = {.5,.75,0,.25},
	[4] = {.75,1,0,.25},
	[5] = {0,.25,.25,.5},
	[6] = {.25,.5,.25,.5},
	[7] = {.5,.75,.25,.5},
	[8] = {.75,1,.25,.5}
}

function ClearAllFrames()
  for i=1, 64 do
    frame=_G["PvEFrame"..i]
    button=_G["PvEFrameButton"..i]
    frame.health.bar:SetValue(0)
    frame.health.bar:SetStatusBarColor(.19,.19,.19)
    frame.health.bar.name:SetText("")
    frame.health.bar.icon:Hide()
    frame.cast.bar:SetValue(0)
    frame.cast.bar.icon:Hide()
    button:SetAttribute("macrotext1","")
  end
end













--Okay. So here is the code behind actually detecting and maintaining data on the mobs.

function NewMob(self, event, arg1)
  local current=tonumber(MOB_DATA["current"])
  local unit=(arg1.."-target")
  if UnitExists(unit) then
    local name=UnitName(unit)
    local guid=UnitGUID(unit)
    local B = tonumber(guid:sub(5,5), 16)
    local maskedB = B % 8
    local reaction=UnitReaction("player",unit)
    local creatureType = UnitCreatureType(unit)
    local raidicon = GetRaidTargetIndex(unit)
    local level=UnitLevel(unit)
    local health=UnitHealth(unit)
    local healthmax=UnitHealthMax(unit)
    local dead = UnitIsDead(unit)
    if current<64 and maskedB==3 and reaction<5 and creatureType~="Critter" and not dead then  --Less than 64 mobs, and NPC, and neutral/lower, non-critter, alive
      --At this point the unit is eligible to become a new target. It must be checked that it is not currently a target.
      local eligible=1
      for i=1, current do
        if guid==MOB_DATA[i]["guid"] then
          eligible=0
        end
      end
      if eligible==1 then
        local frame=_G["PvEFrame"..(current+1)]
        local button=_G["PvEFrameButton"..(current+1)]
        MOB_DATA[(current+1)]["guid"]=guid
        MOB_DATA[(current+1)]["name"]=name
        MOB_DATA[(current+1)]["level"]=level
        MOB_DATA[(current+1)]["health"]=health
        MOB_DATA[(current+1)]["healthmax"]=healthmax
        MOB_DATA[(current+1)]["raidicon"]=raidicon
        frame:Show()
        SetName(frame,name)
        if raidicon then
          SetRaidIcon(frame,raidicon)
        end
        SetHealthBarValues(frame,health,healthmax)
        if InCombatLockdown()==nil then
          button:SetAttribute("macrotext1","/targetexact "..name)
        end
        MOB_DATA["current"]=current+1
      end
    end
  end
end

function PvEUpdate_OnEvent(self, event, ...)
  local timestamp, type, hideCaster, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags = select(1, ...)
  --First need to identify if the source is a monitored NPC.
  local number
  for i=1, 64 do
    if sourceGUID==MOB_DATA[i]["guid"] then
      number=i
    end
  end
  if number~=nil and type=="SPELL_CAST_START" then
    local frame=_G["PvEFrame"..number]
    local spellId, spellName, spellSchool = select(12, ...)
    local _, _, icon, _, _, _, castTime = GetSpellInfo(spellId)
    frame.cast.bar:SetMinMaxValues(0,castTime)
    frame.cast.bar.icon:SetTexture(icon)
    frame.cast.bar.icon:Show()animationGroup=frame:CreateAnimationGroup()
    frame.animationGroup=animationGroup;
    local animation=frame.animationGroup:CreateAnimation("Translation");
    animation:SetOffset(0,0);
    animation:SetDuration(castTime/1000);
    animation:SetScript("OnUpdate",SetCastBarProgress)
    animation:SetScript("OnFinished",SetCastBarZero)
    frame.animationGroup.animation=animation;
    animationGroup:Play()
  end
  if number~=nil and type=="SPELL_CAST_FAILED" then
    local frame=_G["PvEFrame"..number]
    frame.animationGroup.animation:Stop()
    SetCastBarZero(frame.animationGroup.animation)
  end
end



function SetCastBarProgress(self)
  frame=self:GetParent():GetParent()
  frame.cast.bar:SetValue(frame.animationGroup.animation:GetProgress()*frame.animationGroup.animation:GetDuration()*1000)
end

function SetCastBarZero(self)
  frame=self:GetParent():GetParent()
  frame.cast.bar:SetValue(0)
  frame.cast.bar.icon:Hide()
end




function PvEUpdate_OnUpdate()
  --The goal is to check every possible known npc id to gather information about the monitored mobs. This includes target-of-target, mouseover, and focus.
  for i=1, 64 do
    local frame=_G["PvEFrame"..i]
    if UnitGUID("target")==MOB_DATA[i]["guid"] and UnitExists("target") then
      SetBorderColor(frame,1)
    else
      SetBorderColor(frame,0.23828125)
    end
  end
  for a=1, #(PVEFRAMES_UPDATE_UNITS) do
    local unit=PVEFRAMES_UPDATE_UNITS[a]
    if UnitExists(unit) then
      for i=1, 64 do
        if UnitGUID(unit)==MOB_DATA[i]["guid"] then
          local frame=_G["PvEFrame"..i]
          MOB_DATA[i]["name"]=UnitName(unit)
          MOB_DATA[i]["level"]=UnitLevel(unit)
          MOB_DATA[i]["health"]=UnitHealth(unit)
          MOB_DATA[i]["healthmax"]=UnitHealthMax(unit)
          MOB_DATA[i]["raidicon"]=GetRaidTargetIndex(unit)
          if not UnitIsDead(unit) then
            SetName(frame,UnitName(unit))
            if GetRaidTargetIndex(unit) then
              SetRaidIcon(frame,GetRaidTargetIndex(unit))
            end
            SetHealthBarValues(frame,UnitHealth(unit),UnitHealthMax(unit))
            SetHealthBarColor(frame, unit)
          else
            SetName(frame,"")
            frame.health.bar.icon:Hide()
            frame.health.bar:SetValue(0)
            frame.health.bar:SetStatusBarColor(.19,.19,.19)
            frame.cast.bar:SetValue(0)
            frame.cast.bar.icon:Hide()
          end
        end
      end
    end
  end
  for a=1, GetNumPartyMembers() do
    local unit="party"..a.."-target"
    if UnitExists(unit) then
      for i=1, 64 do
        if UnitGUID(unit)==MOB_DATA[i]["guid"] then
          local frame=_G["PvEFrame"..i]
          MOB_DATA[i]["name"]=UnitName(unit)
          MOB_DATA[i]["level"]=UnitLevel(unit)
          MOB_DATA[i]["health"]=UnitHealth(unit)
          MOB_DATA[i]["healthmax"]=UnitHealthMax(unit)
          MOB_DATA[i]["raidicon"]=GetRaidTargetIndex(unit)
          if not UnitIsDead(unit) then
            SetName(frame,UnitName(unit))
            if GetRaidTargetIndex(unit) then
              SetRaidIcon(frame,GetRaidTargetIndex(unit))
            end
            SetHealthBarValues(frame,UnitHealth(unit),UnitHealthMax(unit))
            SetHealthBarColor(frame, unit)
          else
            SetName(frame,"")
            frame.health.bar.icon:Hide()
            frame.health.bar:SetValue(0)
            frame.health.bar:SetStatusBarColor(.19,.19,.19)
            frame.cast.bar:SetValue(0)
            frame.cast.bar.icon:Hide()
          end
        end
      end
    end
  end
  for a=1, GetNumRaidMembers() do
    local unit="raid"..a.."-target"
    if UnitExists(unit) then
      for i=1, 64 do
        if UnitGUID(unit)==MOB_DATA[i]["guid"] then
          local frame=_G["PvEFrame"..i]
          MOB_DATA[i]["name"]=UnitName(unit)
          MOB_DATA[i]["level"]=UnitLevel(unit)
          MOB_DATA[i]["health"]=UnitHealth(unit)
          MOB_DATA[i]["healthmax"]=UnitHealthMax(unit)
          MOB_DATA[i]["raidicon"]=GetRaidTargetIndex(unit)
          if not UnitIsDead(unit) then
            SetName(frame,UnitName(unit))
            if GetRaidTargetIndex(unit) then
              SetRaidIcon(frame,GetRaidTargetIndex(unit))
            end
            SetHealthBarValues(frame,UnitHealth(unit),UnitHealthMax(unit))
            SetHealthBarColor(frame, unit)
          else
            SetName(frame,"")
            frame.health.bar.icon:Hide()
            frame.health.bar:SetValue(0)
            frame.health.bar:SetStatusBarColor(.19,.19,.19)
            frame.cast.bar:SetValue(0)
            frame.cast.bar.icon:Hide()
          end
        end
      end
    end
  end
end

PVEFRAMES_UPDATE_UNITS = {
	[1] = "target",
	[2] = "target-target",
	[3] = "target-target-target",
	[4] = "mouseover",
	[5] = "mouseover-target",
	[6] = "mouseover-target-target",
	[7] = "focus",
	[8] = "focus-target",
	[9] = "focus-target-target",
}

function ClearMobCache()
  for i=1, 64 do
    MOB_DATA[i]={}
  end
  MOB_DATA["current"]=0
end

function ResetFrames()
  ClearMobCache()
  ClearAllFrames()
end

MOB_DATA = {}

MobDetector = CreateFrame("frame")
MobDetector:RegisterEvent("UNIT_TARGET")
MobDetector:SetScript("OnEvent", NewMob)
ClearMobCache()

MobUpdater = CreateFrame("frame")
MobUpdater:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
MobUpdater:SetScript("OnEvent",PvEUpdate_OnEvent)
MobUpdater:SetScript("OnUpdate",PvEUpdate_OnUpdate)

PvEFrameReset=CreateFrame("frame")
PvEFrameReset:RegisterEvent("PLAYER_REGEN_ENABLED")
PvEFrameReset:SetScript("OnEvent",ResetFrames)









--Common words attached to a mob's name that can be left out.

function ShortenName(name)
  for i=1, #(NAME_SHORTENING_LIST) do
    name = gsub(name,NAME_SHORTENING_LIST[i].." ","")
  end
  return name
end

NAME_SHORTENING_LIST = {
	--Throne of the Tides
	[1] = "Commander",
	[2] = "Lady",
	[3] = "Mindbender",
	[4] = "Deep Murloc",
	[5] = "Ebon",
	[6] = "Faceless",
	[7] = "Gilgoblin",
	[8] = "Legionnaire",
	[9] = "Naz'jar",
	[10] = "Unstable",
	[11] = "Unyielding",
	[12] = "Vicious",
	--Blackrock Caverns
	[13] = "Ascendant",
	[14] = "Crazed",
	[15] = "Defiled",
	[16] = "Evolved",
	[17] = "Twilight",
	--Grim Batol
	[18] = "Forgemaster",
	[19] = "Ascended",
	[20] = "Azureborne",
	[21] = "Battered",
	[22] = "Crimsonborne",
	[23] = "Enslaved",
	[24] = "Invoked",
	[25] = "Malignant",
	--Halls of Origination
	[26] = "Temple Guardian",
	[27] = "Blistering",
	[28] = "Bloodpetal",
	[29] = "Celestial",
	[30] = "Dustbone",
	[31] = "Lifewarden",
	[32] = "Stone Trogg",
	[33] = "Sun-Touched",
	[34] = "Temple",
	[35] = "Venomous",
	[36] = "Void",
	--Lost City of the Tol'vir
	[37] = "High",
	[38] = "Bonesnapper",
	[39] = "Frenzied",
	[40] = "Neferset",
	[41] = "Oathsworn",
	[42] = "Pygmy",
	[43] = "Ramkahen",
	[44] = "Tamed",
	[45] = "Tol'vir",
	[46] = "Venomfang",
	--The Stonecore
	[47] = "Crimson",
	[48] = "Crystal",
	[49] = "Crystalspawn",
	[50] = "Devout",
	[51] = "FILLER FOR A SKIPPED NUMBER",
	[52] = "Rock",
	[53] = "Stonecore",
	[54] = "Shale",
	[55] = "Unbound",
	--The Vortex Pinnacle
	[56] = "Armored",
	[57] = "Young",
	--Zul'Aman
	[58] = "Amani",
	[59] = "Amani'shi",
	[60] = "Blood",
	[61] = "Witch Doctor",
	[62] = "Zandalari",
	--Zul'Gurub
	[63] = "Berserking",
	[64] = "Blackwater",
	[65] = "Bloodlord",
	[66] = "Bloodslayer",
	[67] = "Curious",
	[68] = "Florawing",
	[69] = "Gurubashi",
	[70] = "High",
	[71] = "Lesser",
	[72] = "Venomguard",
	[73] = "Venomtip",
	[74] = "Zanzili",


	--Firelands
	[75] = "Ancient",
	[76] = "Blazing",
	[77] = "Cinderweb",
	[78] = "Engorged",
	[79] = "Fire",
	[80] = "Flamewaker",
	[81] = "Lava",
	[82] = "Molten",
	--Azjol Nerub (my testing instance)
	[83] = "Anub'ar",
	[84] = "Skittering",
	--Old Kingdom (test site #2)
	[85] = "Ahn'kahar",
	[86] = "Bound",
	[87] = "Plundering",
	[88] = "Savage",
	[89] = "Twilight",
	--Outside testing
	[90] = "Mature",
	[91] = "Bloodtalon"
}









--Sorting.  The idea is to swap data from frame to frame when out of combat to organize the units. Raid icons first (1-8) then alphabetical order.

function PvEFrames_Sorting()
  local tableOfNames = {} 
  for i=1, (MOB_DATA["current"]) do
    local name, guid = MOB_DATA[i]["name"], MOB_DATA[i]["guid"]
    tableOfNames[name] = guid
  end
  sort(tableOfNames)
  for index,value in pairs(tableOfNames) do
    DEFAULT_CHAT_FRAME:AddMessage(index.." : "..value)
  end
end

function SwapInformation(id1,id2)
  local temporary = {}
  temporary["guid"], temporary["name"], temporary["level"], temporary["health"], temporary["healthmax"], temporary["raidicon"] = MOB_DATA[id1]["guid"], MOB_DATA[id1]["name"], MOB_DATA[id1]["level"], MOB_DATA[id1]["health"], MOB_DATA[id1]["healthmax"], MOB_DATA[id1]["raidicon"]
  MOB_DATA[id1]["guid"], MOB_DATA[id1]["name"], MOB_DATA[id1]["level"], MOB_DATA[id1]["health"], MOB_DATA[id1]["healthmax"], MOB_DATA[id1]["raidicon"] = MOB_DATA[id2]["guid"], MOB_DATA[id2]["name"], MOB_DATA[id2]["level"], MOB_DATA[id2]["health"], MOB_DATA[id2]["healthmax"], MOB_DATA[id2]["raidicon"]
  MOB_DATA[id2]["guid"], MOB_DATA[id2]["name"], MOB_DATA[id2]["level"], MOB_DATA[id2]["health"], MOB_DATA[id2]["healthmax"], MOB_DATA[id2]["raidicon"] = temporary["guid"], temporary["name"], temporary["level"], temporary["health"], temporary["healthmax"], temporary["raidicon"]
  _G["PvEFrameButton"..id1]:SetAttribute("macrotext1","/targetexact "..MOB_DATA[id1]["name"])
  _G["PvEFrameButton"..id2]:SetAttribute("macrotext1","/targetexact "..MOB_DATA[id2]["name"])
end